#!/usr/bin/python
#
# @name = 'storage.py'
#
# @description = "Controller of storage modules."
#
# @author = ['YU Chiu Man', 'HU Yuchong', 'TANG Yang']
#


import Queue
import threading
import os
import gzip

import common
import storages
from storages import *

storagelist = storages.__all__

storageQueue = Queue.Queue()

class StorageThread(threading.Thread):
    '''Class of storage-related threads.'''
    def __init__(self, queue):
        threading.Thread.__init__(self)
        self.queue = queue

    def run(self):
        while True:
            method, args = self.queue.get()
            getattr(self, method)(*args)
            self.queue.task_done()

    def syncMirror(self, setting, nodeid, path):
        '''Sync file entries from cloud to file system.'''
        nodetype = setting.nodeInfo[nodeid].nodetype
        if nodetype in storagelist:
            module = globals()[nodetype]
            func = getattr(module, 'syncMirror')
            return func(setting, nodeid, path)
        else:
            return False

    def uploadFile(self, setting, nodeid, path, dest):
        '''Upload a file to cloud.'''
        nodetype = setting.nodeInfo[nodeid].nodetype
        if nodetype in storagelist:
            module = globals()[nodetype]
            func = getattr(module, 'uploadFile')
            return func(setting, nodeid, path, dest)
        else:
            return False

    def uploadMetadata(self, setting, nodeid, metadatapath, dest):
        '''Upload metadata to cloud.'''
        nodetype = setting.nodeInfo[nodeid].nodetype
        if nodetype in storagelist:
            module = globals()[nodetype]
            func = getattr(module, 'uploadMetadata')
            return func(setting, nodeid, metadatapath, dest)
        else:
            return False

    def uploadFileAndMetadata(self, setting, nodeid, path, metadatapath, dest):
        '''Upload file and metadata to cloud.'''
        nodetype = setting.nodeInfo[nodeid].nodetype
        if nodetype in storagelist:
            module = globals()[nodetype]
            func = getattr(module, 'uploadFileAndMetadata')
            return func(setting, nodeid, path, metadatapath, dest)
        else:
            return False

    def downloadFile(self, setting, nodeid, path, dest):
        '''Download an object from cloud.'''
        nodetype = setting.nodeInfo[nodeid].nodetype
        if nodetype in storagelist:
            module = globals()[nodetype]
            func = getattr(module, 'downloadFile')
            return func(setting, nodeid, path, dest)
        else:
            return False

    def partialDownloadFile(self, setting, nodeid, path, dest, offset, chunksize):
        '''Partial download an object from cloud.'''
        nodetype = setting.nodeInfo[nodeid].nodetype
        if nodetype in storagelist:
            module = globals()[nodetype]
            func = getattr(module, 'partialDownloadFile')
            return func(setting, nodeid, path, dest, offset, chunksize)
        else:
            return False


def initStorages(setting):
    '''Spawn a pool of threads, and pass them queue instance.'''
    for i in range(setting.storagethreads):
        t = StorageThread(storageQueue)
        t.daemon = True
        t.start()


def resetStorageQueue():
    '''Re-initialize the storage queue.'''
    storageQueue.__init__()


def waitForStorages():
    '''Wait on the queue until everything has been processed.'''
    storageQueue.join()


def checkHealth(setting):
    '''Check healthiness of nodes and update setting.'''
    setting.healthynode = setting.totalnode
    for i in range(setting.totalnode):
        nodetype = setting.nodeInfo[i].nodetype
        if nodetype in storagelist:
            module = globals()[nodetype]
            func = getattr(module, 'checkHealth')
            health = func(setting, i)
            setting.nodeInfo[i].healthy = health
            if health == False:
                setting.healthynode -= 1


def syncMirror(setting, path):
    '''Synchronize file entries on cloud to local directory.'''
    for i in range(setting.totalnode):
        if setting.nodeInfo[i].healthy == True:
            nodeid = i
            storageQueue.put(('syncMirror', (setting, nodeid, path)))
    waitForStorages()


def uploadFileAndMetadata(setting, metadata, mode):
    '''Upload file and the metadata to cloud.'''
    checkHealth(setting)
    if setting.healthynode == setting.totalnode:
        #Generate metadata file:
        configname = metadata.filename + ".metadata"
        configpath = setting.metadatadir + '/' + configname
        metadata.write(configpath)

        #Upload big-chunks and metadata:
        for i in range(metadata.totalnode):
            if metadata.fileNodeInfo[i].action == 'upload':
                #Upload big-chunk:
                storageQueue.put(('uploadFileAndMetadata', \
                    (setting, i, metadata.fileNodeInfo[i].bigchunkpath, \
                     configpath, metadata.fileNodeInfo[i].bigchunkname)))
            else:
                #Upload metadata only:
                storageQueue.put(('uploadMetadata', \
                    (setting, i, configpath, \
                     metadata.filename + ".node%d" % i)))
        waitForStorages()
        print "*** upload file finished: " + metadata.filename
    return True


def deleteFile(setting, nodeid, name):
    '''Delete an object on cloud.'''
    nodetype = setting.nodeInfo[nodeid].nodetype
    if nodetype in storagelist:
        module = globals()[nodetype]
        func = getattr(module, 'deleteFile')
        return func(setting, nodeid, name)
    else:
        return False


def downloadMetadata(setting, metadata):
    '''Download metadata file and get metadata.'''
    checkHealth(setting)
    configpath = setting.metadatadir + '/' + metadata.filename + ".metadata"
    retstat = False
    #Download metadata file from cloud:
    for i in range(setting.totalnode):
        if setting.nodeInfo[i].healthy == True:
            nodetype = setting.nodeInfo[i].nodetype
            if nodetype in storagelist:
                module = globals()[nodetype]
                func = getattr(module, 'downloadMetadata')
                if func(setting, i, metadata.filename + ".node%d" % i, configpath) == True:
                    retstat = True
                    break

    if retstat == True:
        metadata.read(configpath, setting)
    return retstat


def downloadFile(setting, metadata):
    '''Download an object from cloud.'''
    checkHealth(setting)
    for i in range(metadata.totalchunk):
        if metadata.chunkInfo[i].action == 'download':
            nodeid = metadata.chunkInfo[i].nodeid
            if metadata.fileNodeInfo[nodeid].chunknum == 1:
                #For one chunk per node:
                chunkpath = metadata.chunkInfo[i].chunkpath
                dest = setting.chunkdir + '/' + chunkpath
                storageQueue.put(('downloadFile', \
                    (setting, nodeid, chunkpath, dest)))
            elif metadata.fileNodeInfo[nodeid].chunknum > 1:
                #For multiple chunks per node:
                bigchunkpath = metadata.fileNodeInfo[nodeid].bigchunkpath
                dest = setting.chunkdir + '/' + metadata.chunkInfo[i].chunkpath
                chunksize = metadata.chunkInfo[i].chunksize
                offset = chunksize * metadata.chunkInfo[i].position
                storageQueue.put(('partialDownloadFile', \
                    (setting, nodeid, bigchunkpath, dest, offset, chunksize)))
            else:
                #For invalid chunk number:
                return False
            #Update metadata:
            metadata.chunkInfo[i].chunkpath = dest
            metadata.chunkInfo[i].healthy = True
    waitForStorages()

    return True


def deleteChunkAndMetadata(setting, metadata):
    '''Delete chunks and metadata of a file on cloud.'''
    #Delete chunks:
    for fileNode in metadata.fileNodeInfo:
        try:
            deleteFile(setting, fileNode.nodeid, fileNode.bigchunkpath)
        except:
            return False
    return True
